# -*- coding: utf-8 -*-
from __future__ import unicode_literals, print_function, division, absolute_import
from bgfiles import toolbox
from bgfiles.patterns import registry
from django.test import TestCase
from django.utils import timezone


class SimpleHandler(object):
    handled = []

    @classmethod
    def handle(cls, request):
        cls.handled.append(request)


class RegistryTest(TestCase):

    def setUp(self):
        self.original_handlers = registry.handlers.copy()
        registry.handlers.clear()
        SimpleHandler.handled = []

    def tearDown(self):
        registry.handlers.clear()
        registry.handlers.update(self.original_handlers)

    def test(self):
        registry.register(SimpleHandler, ['something'])
        request = toolbox.add_request('abc', file_type='something')
        self.assertTrue(registry.dispatch(request.id))
        self.assertEqual([request], SimpleHandler.handled)

    def test_no_handler(self):
        request = toolbox.add_request('abc')
        self.assertRaises(KeyError, registry.dispatch, request.id)
        registry.register(SimpleHandler, ['something'])
        self.assertRaises(KeyError, registry.dispatch, request.id)

    def test_already_finished(self):
        registry.register(SimpleHandler, ['something'])
        request = toolbox.add_request('abc', file_type='something')
        request.finished_at = timezone.now()
        request.save()
        self.assertFalse(registry.dispatch(request.id))
        self.assertEqual(0, len(SimpleHandler.handled))



# -*- coding: utf-8 -*-
from __future__ import unicode_literals, print_function, division, absolute_import
from django.conf import settings
from django.db.models import FileField


def get_storage():
    """Get the configured storage.

    :return: the configured storage or None, to force the use of Django's default storage
    """
    return getattr(settings, 'BG_FILES_STORAGE', None)


def get_upload_to():
    """Get the value for FileRequest's ``upload_to``.

    :return: the configured value or an empty string
    """
    return getattr(settings, 'BG_FILES_UPLOAD_TO', '')


class ConfigurableFileField(FileField):
    """Extension of Django's regular model FileField to prevent migration generation for every single change.

    Because the FileField will include the storage and upload_to attributes in the ``deconstruct``, this means
    any change to the relevant settings will cause Django to complain about missing migrations.

    If you want to have a copy of the old values for storage and upload_to in your migrations, please provide them
    yourself.

    """

    def __init__(self, *args, **kwargs):
        kwargs.setdefault('storage', get_storage())
        kwargs.setdefault('upload_to', get_upload_to())
        super(ConfigurableFileField, self).__init__(*args, **kwargs)

    def deconstruct(self):
        name, path, args, kwargs = super(ConfigurableFileField, self).deconstruct()
        kwargs.pop('storage', None)
        kwargs.pop('upload_to', None)
        return name, path, args, kwargs



# -*- coding: utf-8 -*-
from bgfiles.fields import get_storage, get_upload_to
from django.test import SimpleTestCase, override_settings


class FileRequestSettingsTest(SimpleTestCase):

    def test_get_storage_default(self):
        self.assertTrue(get_storage() is None)

    def test_get_storage_overridden(self):
        sentinel = object()
        with override_settings(BG_FILES_STORAGE=sentinel):
            self.assertTrue(get_storage() is sentinel)

    def test_get_upload_to_default(self):
        self.assertEqual('', get_upload_to())

    def test_get_upload_to_overridden(self):
        upload_dir = '/oh/my/temp/'
        with override_settings(BG_FILES_UPLOAD_TO=upload_dir):
            self.assertEqual(upload_dir, get_upload_to())



# -*- coding: utf-8 -*-
from __future__ import unicode_literals, absolute_import
from django.db import models
from django.conf import settings
from django.utils import timezone
from django.utils.encoding import python_2_unicode_compatible
from uuid import uuid4
import time
from .fields import ConfigurableFileField


class FileRequestManager(models.Manager):

    def to_handle(self):
        """Grab requests that still need to be handled.

        The requests are ordered by their date of requests, so older requests take precedence.

        :return: request queryset
        """
        return self.filter(finished_at__isnull=True).order_by('requested_at')

    def remove_expired(self, treshold, timeout=None, sleep=None, max_items=None):
        """Removes expired requests.

        It will return a three-tuple:

        - status: one of "processed", "max_items" or "timeout", depending on why the processing stopped
        - number of requests deleted
        - total number of requests that were available to delete

        :param treshold: requests should have expired before this
        :type treshold: datetime
        :param timeout: epoch time specifying when to stop processing
        :type timeout: int
        :param sleep: number of seconds to sleep in between requests
        :type sleep: float
        :param max_items: maximum number of items to remove
        :type max_items: int

        :return: (status, nr deleted, total items available to delete)
        """
        requests = self.filter(expires_at__lte=treshold)
        total = requests.count()
        nr_deleted = 0
        status = 'processed'
        for download in requests.order_by('expires_at').iterator():
            if sleep and nr_deleted > 0:
                time.sleep(sleep)
            # We process them manually to ensure the delete method is called, deleting any attached files as well.
            download.delete(using=self._db)
            nr_deleted += 1
            if max_items and nr_deleted == max_items:
                status = 'max_items'
                break
            if timeout and time.time() > timeout:
                status = 'timeout'
                break
        return status, nr_deleted, total


@python_2_unicode_compatible
class FileRequest(models.Model):
    id = models.UUIDField(primary_key=True, default=uuid4)
    criteria = models.TextField(blank=True)
    filehandle = ConfigurableFileField(blank=True, null=True)
    filename = models.CharField(max_length=200, blank=True)
    description = models.TextField(blank=True)
    file_type = models.CharField(max_length=100, blank=True, db_index=True)
    content_type = models.CharField(max_length=200, blank=True)
    size = models.BigIntegerField(blank=True, null=True)
    requester = models.ForeignKey(settings.AUTH_USER_MODEL, blank=True, null=True, related_name='requested_downloads',
                                  on_delete=models.CASCADE)
    requester_language = models.CharField(max_length=10, blank=True)
    requester_timezone = models.CharField(max_length=200, blank=True)
    requested_at = models.DateTimeField(db_index=True)
    finished_at = models.DateTimeField(blank=True, null=True, db_index=True)
    expires_at = models.DateTimeField(blank=True, null=True)
    objects = FileRequestManager()

    class Meta:
        ordering = ['-requested_at']

    def __str__(self):
        return '{} requested at {} by {}'.format(self.file_type, self.requested_at, self.requester)

    def delete(self, *args, **kwargs):
        # Ensure we remove our file
        if self.filehandle:
            self.filehandle.delete(save=False)
        return super(FileRequest, self).delete(*args, **kwargs)

    def attach_file(self, contents, filename=None, content_type=None, size=None, timestamp=None):
        timestamp = timestamp if timestamp else timezone.now()
        if filename:
            self.filename = filename
        if content_type:
            self.content_type = content_type
        if size:
            self.size = size
        self.filehandle.save(self.filename, contents)
        self.finished_at = timestamp

    def expire_in(self, delta, timestamp=None):
        timestamp = timestamp if timestamp else timezone.now()
        self.expires_at = timestamp + delta



# -*- coding: utf-8 -*-
from __future__ import unicode_literals, print_function, division, absolute_import
from bgfiles import toolbox
from bgfiles.models import FileRequest
from django.conf import settings
from django.contrib.auth.models import User
from django.core.files.base import ContentFile
from django.core.signing import BadSignature
from django.http import Http404
from django.utils import six
from django.test import SimpleTestCase, TestCase, override_settings
import os
import tempfile
from .utils import WithTempDir


class GetDefaultSignerTest(SimpleTestCase):

    def test(self):
        signer = toolbox.get_default_signer()
        self.assertEqual(signer.key, settings.SECRET_KEY)
        self.assertEqual(signer.salt, toolbox.Signer.DEFAULT_SALT)
        self.assertEqual(signer.serializer, toolbox.Signer.DEFAULT_SERIALIZER)
        self.assertEqual(signer.max_age, 86400)

    def test_overridden(self):
        signer = toolbox.Signer()
        with override_settings(BGFILES_DEFAULT_SIGNER=signer):
            self.assertTrue(toolbox.get_default_signer() is signer)

    def test_overridden_settings(self):
        my_key = 'abclkdjsldkjsldkjsd__*((#*))'
        my_salt = 'wiouoegwgr'
        my_serializer_class = type(str('MySerializer'), (toolbox.JSONSerializer,), {})
        my_max_age = 3540
        kwargs = {
            'BGFILES_DEFAULT_KEY': my_key,
            'BGFILES_DEFAULT_SALT': my_salt,
            'BGFILES_DEFAULT_SERIALIZER': my_serializer_class,
            'BGFILES_DEFAULT_MAX_AGE': my_max_age
        }
        with override_settings(**kwargs):
            signer = toolbox.get_default_signer()
        self.assertEqual(signer.key, my_key)
        self.assertEqual(signer.salt, my_salt)
        self.assertEqual(signer.serializer, my_serializer_class)
        self.assertEqual(signer.max_age, my_max_age)


class SignerLoadsAndDumpsTest(SimpleTestCase):

    def test(self):
        signer = toolbox.get_default_signer()
        data_in = {'a': 1, 'b': 3, 'www': {'abc': ['y', 'z']}}
        token = signer.dump(data_in)
        data_out = signer.load(token)
        self.assertDictEqual(data_in, data_out)


class AddRequestTest(TestCase):

    def test(self):
        criteria = 'abc'
        request = toolbox.add_request(criteria)
        self.assertEqual(criteria, request.criteria)
        self.assertTrue(request.requested_at)
        self.assertFalse(request.file_type)
        self.assertFalse(request.requester)
        self.assertFalse(request.content_type)
        self.assertTrue(request.id)
        content_type = 'text/csv'
        request = toolbox.add_request(criteria, content_type=content_type)
        self.assertEqual(criteria, request.criteria)
        self.assertEqual(content_type, request.content_type)
        self.assertTrue(request.requested_at)
        self.assertFalse(request.file_type)
        self.assertFalse(request.requester)
        file_type = 'some-report'
        request = toolbox.add_request(criteria, content_type=content_type, file_type=file_type)
        self.assertEqual(criteria, request.criteria)
        self.assertEqual(content_type, request.content_type)
        self.assertEqual(file_type, request.file_type)
        self.assertTrue(request.requested_at)
        self.assertFalse(request.requester)
        requester = User.objects.create_user('something', 'mail@example.com', 'abc')
        request = toolbox.add_request(criteria, content_type=content_type, file_type=file_type, requester=requester)
        self.assertEqual(criteria, request.criteria)
        self.assertEqual(content_type, request.content_type)
        self.assertEqual(file_type, request.file_type)
        self.assertEqual(requester, request.requester)
        self.assertTrue(request.requested_at)


class AttachFileTest(WithTempDir, TestCase):

    def test(self):
        request = toolbox.add_request('abc')
        with override_settings(MEDIA_ROOT=self.tempdir, MEDIA_URL=''):
            contents = ContentFile('something special')
            toolbox.attach_file(request, contents, 'something.txt', 'text/plain')
            path = request.filehandle.path
            self.assertTrue(os.path.isfile(path))
            request.delete()
            self.assertFalse(os.path.isfile(path))

    def test_attach_local_file(self):
        fh, temppath = tempfile.mkstemp('test')
        os.close(fh)
        with open(temppath, 'w') as fh:
            fh.write(six.text_type('something special'))
        request = toolbox.add_request('abc')
        with override_settings(MEDIA_ROOT=self.tempdir, MEDIA_URL=''):
            toolbox.attach_local_file(request, temppath, 'something.txt', 'text/plain')
            path = request.filehandle.path
            self.assertTrue(os.path.isfile(path))
            request.delete()
            self.assertFalse(os.path.isfile(path))


class CreateTokenTest(TestCase):

    def test(self):
        requester = User.objects.create_user('ownow', 'mail@example.com', 'mypassword')
        request = toolbox.add_request('abc', requester=requester)
        token = toolbox.create_token(request)
        data = toolbox.get_default_signer().load(token)
        self.assertEquals(data.get('id'), '%s' % request.id)
        self.assertEquals(data.get('requester'), requester.id)
        # Now with additional data
        additional_data = {'abc': 123, 'z': ['y', 'z', 1]}
        token = toolbox.create_token(request, data=additional_data)
        data = toolbox.get_default_signer().load(token)
        additional_data['id'] = '%s' % request.id
        additional_data['requester'] = requester.id
        self.assertDictEqual(data, additional_data)

    def test_missing_requester(self):
        request = toolbox.add_request('abc')
        self.assertRaises(RuntimeError, toolbox.create_token, request)

    def test_general(self):
        requester = User.objects.create_user('ownow', 'mail@example.com', 'mypassword')
        request = toolbox.add_request('abc', requester=requester)
        token = toolbox.create_general_token(request)
        data = toolbox.get_default_signer().load(token)
        self.assertEquals(data.get('id'), '%s' % request.id)
        self.assertFalse('requester' in data)
        # Now with additional data
        additional_data = {'abc': 123, 'z': ['y', 'z', 1]}
        token = toolbox.create_general_token(request, data=additional_data)
        data = toolbox.get_default_signer().load(token)
        additional_data['id'] = '%s' % request.id
        self.assertDictEqual(data, additional_data)

    def test_general_does_not_require_requester(self):
        request = toolbox.add_request('abc')
        toolbox.create_general_token(request)


class DecodeTest(TestCase):

    def test(self):
        requester = User.objects.create_user('ownow', 'mail@example.com', 'mypassword')
        request = toolbox.add_request('abc', requester=requester)
        token = toolbox.create_token(request)
        # Clear the requester
        request.requester = None
        request.save()
        self.assertRaises(FileRequest.DoesNotExist, toolbox.decode, token, require_requester=True)
        request.requester = requester
        request.save()
        found, data = toolbox.decode(token, require_requester=True)
        self.assertEqual(found, request)

    def test_general(self):
        request = toolbox.add_request('abc')
        token = toolbox.create_general_token(request)
        request.delete()
        self.assertRaises(FileRequest.DoesNotExist, toolbox.decode, token, require_requester=False)

    def test_requester_required_but_not_present(self):
        request = toolbox.add_request('abc')
        token = toolbox.create_general_token(request)
        request.delete()
        self.assertRaises(BadSignature, toolbox.decode, token, require_requester=True)


class ServeTest(WithTempDir, TestCase):

    def test_no_file_yet(self):
        request = toolbox.add_request('abc')
        self.assertRaises(Http404, toolbox.serve, request)
        to_raise = ValueError
        self.assertRaises(to_raise, toolbox.serve, request, raise_on_missing=to_raise)

    def test(self):
        request = toolbox.add_request('abc')
        with override_settings(MEDIA_ROOT=self.tempdir, MEDIA_URL=''):
            contents = ContentFile('something special')
            toolbox.attach_file(request, contents, 'something.txt')
            response = toolbox.serve(request)
            self.assertEqual(response['Content-Type'], 'application/octet-stream')
            self.assertEqual(response['Content-Disposition'], 'attachment; filename="something.txt"')
            self.assertEqual(response.content, b'something special')
            response = toolbox.serve(request, default_content_type='text/plain')
            self.assertEqual(response['Content-Type'], 'text/plain')
            self.assertEqual(response.content, b'something special')
            self.assertEqual(response['Content-Disposition'], 'attachment; filename="something.txt"')
            request.filename = ''
            request.save()
            response = toolbox.serve(request, default_content_type='text/plain')
            self.assertEqual(response['Content-Type'], 'text/plain')
            self.assertEqual(response.content, b'something special')
            self.assertFalse(response.get('Content-Disposition'))



# -*- coding: utf-8 -*-
from __future__ import unicode_literals, print_function, division, absolute_import

from django.conf import settings
from django.core.files.base import ContentFile
from django.core.mail import send_mail
from django.http import HttpResponse, QueryDict
from django.shortcuts import render
from django.utils.encoding import force_text
from io import BytesIO
from .. import toolbox
from ..http import create_content_disposition


class InvalidCriteria(Exception):
    """Exception to indicate calling code of a pattern that the supplied criteria were invalid."""
    pass


class InvalidForm(InvalidCriteria):
    """Exception to raise when using a form to validate criteria.

    Pass the form as ``form=form`` to allow calling code to display errors.
    """

    def __init__(self, *args, **kwargs):
        form = kwargs.pop('form')
        super(InvalidCriteria, self).__init__(*args, **kwargs)
        self.form = form


class Criteria(object):
    """Wrapper for criteria, allowing marshalling and unmarshalling."""

    def __init__(self, raw):
        self.raw = raw

    def marshall(self):
        """Create a marshalled/serialized version of the criteria.

        This is used to store the criteria on a file request.
        """
        raise NotImplementedError()

    @classmethod
    def restore_from(cls, file_request):
        """Restore a Criteria instance from the file request.

        This is used to restore the criteria for our file generation when loading a file
        request from the database.
        """
        raise NotImplementedError()

    @classmethod
    def build(cls, http_request):
        """Useful method to simply build the initial criteria from an HTTP request.

        :param http_request: http request
        :return: Criteria instance
        """
        raise NotImplementedError()


class QueryDictCriteria(Criteria):
    """Criteria implementation assuming you want to use request.GET or request.POST as criteria.

    The ``raw`` attribute should be an instance of Django's ``QueryDict``.
    """

    def marshall(self):
        """Creates the URL encoded version of our QueryDict after sanitizing.

        :return: string
        """
        return self.sanitize(self.raw).urlencode()

    @classmethod
    def sanitize(cls, raw):
        # TODO Remove csrftoken
        return raw

    @classmethod
    def restore_from(cls, file_request):
        return cls(cls.criteria_query_dict(file_request))

    @classmethod
    def build(cls, http_request):
        return cls(cls.http_query_dict(http_request))

    @classmethod
    def criteria_query_dict(cls, file_request):
        """Create a QueryDict from the file request criteria (i.e. the original query string).

        :param file_request: the file request we want to restore the criteria from
        :return: QueryDict
        """
        encoding = getattr(settings, 'BG_FILES_CRITERIA_ENCODING', None)
        criteria = file_request.criteria
        if encoding:
            criteria = criteria.encode(encoding)
        return QueryDict(query_string=criteria)

    @classmethod
    def http_query_dict(cls, http_request):
        """Grab the correct query dict from the HTTP request.

        :param http_request: request
        :return: QueryDict
        """
        if http_request.method == 'POST':
            return http_request.POST
        if http_request.method == 'GET':
            return http_request.GET
        raise RuntimeError('Default implementation only supports GET and POST')


class QueryDictCriteriaWithRequester(QueryDictCriteria):
    """Extension of QueryDictCriteria that includes the requester.

    Note that the requester won't be serialized in the criteria since it's already available on the file
    request itself.
    """

    def __init__(self, raw, user):
        super(QueryDictCriteriaWithRequester, self).__init__(raw)
        self.user = user

    @classmethod
    def restore_from(cls, file_request):
        return cls(cls.criteria_query_dict(file_request), file_request.requester)

    @classmethod
    def build(cls, http_request):
        return cls(cls.http_query_dict(http_request), http_request.user)


class Dataset(object):
    """Wrapper for the dataset to use when creating the file.

    Contains the criteria instance, items and whether the processing should be delayed
    """

    def __init__(self, criteria, items=None, delay=False):
        self.criteria = criteria
        self.items = items
        self.delay = delay


class SimplePattern(object):
    """The most basic pattern.

    This simply lays out the basic scenario:

    - build the dataset for the current HTTP request
    - if the dataset indicates we need to delay processing, respond accordingly
    - otherwise respond with the file

    """

    def respond_to(self, http_request):
        """Respond to the file creation request.

        Returns a two-tuple (delayed, http response). The first value indicates whether processing was delayed,
        the second value (typically) contains the HTTP response.

        :param http_request: the HTTP request for the file
        :return: two-tuple
        """
        dataset = self.build_dataset(http_request)
        if not dataset.delay:
            return False, self.respond_with_file(http_request, dataset)
        return True, self.respond_with_delay(http_request, dataset)

    def build_dataset(self, http_request):
        """
        Build the dataset to use for this http request.

        :param http_request: the current http request
        :return: a dataset
        """
        raise NotImplementedError()

    def respond_with_file(self, http_request, dataset):
        """Create an HTTP response containing the file.

        :param http_request: the current http request
        :param dataset: the dataset we compiled, so you can re-use the work you've already done
        :return: HTTP response
        """
        raise NotImplementedError()

    def respond_with_delay(self, http_request, dataset):
        """Create a response when we've delayed the file creation.

        :param http_request: the current http request
        :param dataset: the dataset we compiled
        :return: HTTP response
        """
        raise NotImplementedError()


class WriterPattern(SimplePattern):
    """Pattern providing the logic for writing of the file, both online and offline.

    Your implementation should provide:

    - ``get_items``
    - ``evaluate_dataset``
    - ``write_bytes``

    Warning:

    - This pattern will not do a single thing when ``schedule`` is called. It assumes you use the information in
      the file request to process them with for example a cron job.
    - This pattern will not send out a notification. Implement ``send_notification`` to do so.
    - This pattern will not create an HTTP response when the processing is delayed. You'll have to implement
      ``create_delayed_response`` to do so or let your view handle it.

    See FullPattern for a pattern providing these options.

    """

    # The criteria class to use; defaults to QueryDictCriteriaWithRequester
    criteria_class = QueryDictCriteriaWithRequester
    # The file type to apply to the file request
    file_type = None
    # An optional description of the file to remind the user; stored when adding the file request
    file_description = None
    # The content type for the file
    content_type = None
    # The buffer class to supply to ``write_bytes``; defaults to io.BytesIO
    buffer_class = BytesIO
    # Whether we only want the requester to be able to download the file; defaults to True
    requester_only = True
    # The signer to use; defaults to the configured default signer
    signer = None
    # Whether to force the download of the file when we can create it online; defaults to True
    content_as_attachment = True

    def __init__(self, filename=None):
        self.filename = filename

    def build_dataset(self, http_request):
        """Build the dataset.

        This uses the configured criteria class to wrap the HTTP request, fetches the items using ``get_items``
        and then constructs a dataset to evaluate it with ``evaluate_dataset``.

        :param http_request: the current HTTP request
        :return: dataset
        """
        criteria = self.criteria_class.build(http_request)
        items = self.get_items(criteria)
        dataset = Dataset(criteria, items=items)
        self.evaluate_dataset(dataset, http_request)
        return dataset

    def respond_with_file(self, http_request, dataset):
        """Generate the file and return it as an HTTP response.

        This will simply call the ``write_bytes`` method to generate the file and wrap it in an HTTP response.

        :param http_request: the current HTTP request
        :param dataset: dataset we compiled
        :return: HTTP response
        """
        response = HttpResponse(content_type=self.content_type)
        response['Content-Disposition'] = create_content_disposition(self.get_filename(), self.content_as_attachment)
        bytebuffer = self.get_buffer_class()()
        self.write_bytes(dataset, bytebuffer)
        contents = bytebuffer.getvalue()
        bytebuffer.close()
        response.write(contents)
        return response

    def respond_with_delay(self, http_request, dataset):
        """Schedule a file request and respond to the user.

        This method will call the ``schedule`` method after adding the file request. So if you want custom
        scheduling, that's the place to be.

        :param http_request: the current HTTP request
        :param dataset: dataset we compiled
        :return: HTTP response (or something else)
        """
        file_request = toolbox.add_request(**self.get_file_request_data(http_request, dataset))
        schedule = self.schedule(file_request, http_request, dataset)
        return self.create_delayed_response(file_request, http_request, dataset, schedule)

    def get_items(self, criteria):
        """Retrieve the items we need to use to construct our file.

        Typically this is a queryset, but it might as well be a list, string or whatever you want. It can even
        be nothing at all. Here's an example assuming you're using a form to filter a queryset::

            form = IssueFilterForm(data=criteria.raw)
            issues = Issue.objects.all()
            return form.filter(issues)

        By default this pattern uses the ``QueryDictCriteriaWithRequester`` which means you can utilize the user
        as well::

            form = IssueFilterForm(data=criteria.raw)
            issues = Issue.objects.filter(assignee=criteria.user)
            return form.filter(issues)

        :param criteria: criteria instance defining which items to retrieve
        :return: the items
        """
        raise NotImplementedError()

    def evaluate_dataset(self, dataset, http_request):
        """Evaluate the dataset by setting its ``delay`` attribute.

        The dataset will contain the criteria and the items retrieved using ``get_items``.

        Your implementation might be simple or complex::

            # Simple
            dataset.delay = dataset.items.count > 500
            # Complex
            dataset.delay = CpuMonitor.get_current_load() > 0.25

        :param dataset: the dataset to evaluate
        :param http_request: the current HTTP request
        """
        raise NotImplementedError()

    def schedule(self, file_request, http_request, dataset):
        """Schedule the file request for processing.

        Does nothing by default, assuming you process outstanding requests manually/with a cron job.
        This is where you might schedule a Celery or RQ task for the file request.
        """
        return None

    def write_bytes(self, dataset, output):
        """Write the actual bytes for the file.

        Sorry, can't help you here. This is entirely up to you.

        :param dataset: the dataset containing the items and criteria
        :param output: output to write to; instance of ``get_buffer_class()``
        """
        raise NotImplementedError()

    def create_file(self, file_request):
        """Create the file offline based on the info in the file request.

        This method will:
        - Restore the criteria from the file request
        - Fetch the items using ``get_items``
        - Set up a buffer using ``get_buffer_class``
        - Call ``write_bytes``
        - Attach the corresponding file to the file request
        - Create a token (use ``is_requester_only`` to decide on whether the token is user-specific or general)
        - Pass the token and file request on to ``send_notification``

        :param file_request: request to process
        :return: result of sending out the notification to the user
        """
        criteria = self.criteria_class.restore_from(file_request)
        items = self.get_items(criteria)
        dataset = Dataset(criteria, items)
        bytebuffer = self.get_buffer_class()()
        self.write_bytes(dataset, bytebuffer)
        contents = bytebuffer.getvalue()
        bytebuffer.close()
        toolbox.attach_file(file_request, ContentFile(contents), self.get_filename(), self.get_content_type())
        token_data = self.get_token_data(file_request)
        if self.is_requester_only():
            token = toolbox.create_token(file_request, data=token_data, signer=self.get_signer())
        else:
            token = toolbox.create_general_token(file_request, data=token_data, signer=self.get_signer())
        return self.send_notification(file_request, token)

    def send_notification(self, file_request, token):
        """Send a notification to the user.

        By default this does absolutely nothing. Have a look at FullPattern for one that does something useful.

        :param file_request: the file request
        :param token: the token to access the file
        :return: the result of the notification (e.g. number of mails sent or Celery task)
        """
        return None

    def get_file_request_data(self, http_request, dataset):
        """Set up the data to use when adding a file request.

        :param http_request: the current HTTP request
        :param dataset: the dataset we compiled
        :return: dict
        """
        description = self.get_file_description()
        description = description if description else ''
        return {
            'criteria': dataset.criteria.marshall(),
            'file_type': self.get_file_type(),
            'description': description,
            'requester': http_request.user,
            'content_type': self.get_content_type()
        }

    def get_token_data(self, file_request):
        """Get optional additional data to embed in the token.

        :param file_request: the file request
        :return: dict
        """
        return {}

    def get_content_type(self):
        """Get the content type for the file.

        :return: by default returns the class-level ``content_type``
        """
        return self.content_type

    def get_file_type(self):
        """Get the type of file (e.g. 'report-a' or 'report-b').

        :return: by default returns the class-level ``file_type``
        """
        return self.file_type

    def get_file_description(self):
        """Get the description of the file (e.g. 'Your report A' or 'Your report B').

        :return: by default returns the class-level ``file_description``
        """
        return self.file_description

    def get_filename(self):
        """Get the name of the file (e.g. 'report-a.xlsx' or 'report-b.xlsx').

        :return: by default returns the ``filename`` attribute
        """
        return self.filename

    def is_requester_only(self):
        """Indicates whether we want the token to only be used by the requester of the file.

        :return: defaults to the class-level ``requester_only``.
        """
        return self.requester_only

    def get_buffer_class(self):
        """Get the buffer class to use when writing the bytes.

        :return: defaults to the class-level ``buffer_class``.
        """
        return self.buffer_class

    def get_signer(self):
        """Get the signer to use for our tokens.

        :return: defaults to the class-level ``signer`` or ``None`` (which means use the default one)
        """
        return self.signer if self.signer else None

    def create_delayed_response(self, file_request, http_request, dataset, schedule):
        """Create a response used when the processing is delayed.

        This is where you'd typically provide the user with an informative message to tell them you're working on
        it and you'll let them know when the file's ready.

        :param file_request: the file request we added
        :param http_request: the current HTTP request
        :param dataset: the dataset we compiled
        :param schedule: the result of calling ``schedule``
        :return: typically an HTTP response
        """
        return None

    @classmethod
    def restore(cls, file_request):
        """Restore an instance of this class from a file request.

        :param file_request: file request
        :return: class instance
        """
        return cls(file_request.filename)

    @classmethod
    def handle(cls, file_request):
        """Handle a file request.

        :param file_request: file request
        :return: result of sending out the notification after creating the file
        """
        instance = cls.restore(file_request)
        return instance.create_file(file_request)


class DelayedTemplateResponseMixin(object):
    """Pattern mixin providing a template response when the processing is delayed."""

    # Name of the template to use
    delayed_template_name = None

    def create_delayed_response(self, file_request, http_request, dataset, schedule):
        """Actually creates an HTTP response based on a template and optional context.

        :param file_request: the file request we added
        :param http_request: the current HTTP request
        :param dataset: the dataset we compiled
        :param schedule: the result of scheduling the processing of the HTTP request
        :return: an HTTP response
        """
        context = self.get_delayed_context_data(file_request=file_request, dataset=dataset, schedule=schedule)
        return render(http_request, self.get_delayed_template_name(), context=context)

    def get_delayed_template_name(self):
        """Get the template name to use.

        :return: defaults to the class-level ``delayed_template_name``
        """
        return self.delayed_template_name

    def get_delayed_context_data(self, **kwargs):
        """Hook to provide additional context data.

        :param kwargs: same parameters as ``create_delayed_response``
        :return: dict
        """
        return kwargs


class EmailNotificationMixin(object):
    """Pattern mixin to send out an email when a file is ready for download.

    Example::

        class IssueExport(EmailNotificationMixin, WriterPattern):
            email_subject = _('Your file %(filename)s is ready!')
            email_text_message = _('You can download it here: %(download_url)s')
            download_url = 'https://example.com/secure/%(token)s'

    """

    # Subject(-template) of the email; attributes of the file request and ``download_url`` will be inserted.
    email_subject = None
    # The from address
    from_email = None
    # Text(-template) of the email; attributes of the file request and ``download_url`` will be inserted.
    email_text_message = None
    # HTML(-template) of the email; attributes of the file request and ``download_url`` will be inserted.
    email_html_message = None
    # In case you've got a fixed download url; ``token`` will be inserted.
    download_url = None

    def send_notification(self, file_request, token):
        """Send out an email notification for the finished request.

        This will grab all the components of the email, insert the attributes of the file request and the
        download_url, and send it on its way.

        :param file_request: the file request
        :param token: the token to use
        :return: number of messages sent
        """
        template_kwargs = file_request.__dict__.copy()
        template_kwargs['download_url'] = self.get_download_url(file_request, token)
        subject = force_text(self.get_email_subject()) % template_kwargs
        message = force_text(self.get_email_text_message()) % template_kwargs
        html_message = force_text(self.get_email_html_message()) % template_kwargs
        recipients = self.get_email_recipients(file_request)
        return send_mail(subject, message, self.get_from_email(), recipients, html_message=html_message)

    def get_download_url(self, file_request, token):
        """Construct the download url for the request.

        By default uses the configured ``download_url`` and inserts the token.
        Don't forget to include your host.

        :param file_request: the file request
        :param token: the token to use
        :return: the download url
        """
        template_kwargs = file_request.__dict__.copy()
        template_kwargs['token'] = token
        return force_text(self.download_url) % template_kwargs

    def get_email_recipients(self, file_request):
        """Grab the recipients for the email.

        :param file_request: the file request
        :return: list of recipients; by default uses the requester of the file
        """
        return [file_request.requester.email]

    def get_email_subject(self):
        """Get the subject for the email.

        :return: defaults to the class-level ``email_subject``
        """
        return self.email_subject

    def get_email_text_message(self):
        """Get the text body for the email.

        :return: defaults to the class-level ``email_text_message``
        """
        return self.email_text_message

    def get_email_html_message(self):
        """Get the HTML body for the email.

        :return: defaults to the class-level ``email_html_message``
        """
        return self.email_html_message

    def get_from_email(self):
        """Get the from email address.

        :return: defaults to the class-level ``from_email``
        """
        return self.from_email


class FullPattern(DelayedTemplateResponseMixin, EmailNotificationMixin, WriterPattern):
    """Pattern class incorporating a delayed HTTP response and email notification.

    All that's left to do is implementing:
    - ``evaluate_dataset`` so we know whether to offload the file creation
    - ``write_bytes`` to actually have a file
    - ``get_items`` to select the items (if any) you need to create the file

    """

    def evaluate_dataset(self, dataset, http_request):
        raise NotImplementedError()

    def write_bytes(self, dataset, output):
        raise NotImplementedError()

    def get_items(self, criteria):
        raise NotImplementedError()



# -*- coding: utf-8 -*-
from __future__ import unicode_literals, print_function, division, absolute_import
from bgfiles.models import FileRequest
from bgfiles.patterns import (Dataset, InvalidCriteria, InvalidForm, QueryDictCriteriaWithRequester,
                              QueryDictCriteria, Criteria, SimplePattern)
from django.contrib.auth.models import User
from django.test import RequestFactory, SimpleTestCase


class VerifyExceptionsTest(SimpleTestCase):

    def test(self):
        self.assertTrue(issubclass(InvalidForm, InvalidCriteria))

    def test_invalid_form(self):
        form = object()
        e = InvalidForm('Oh no!', form=form)
        self.assertEqual(e.form, form)


class QueryDictCriteriaWithRequesterTest(SimpleTestCase):

    def test_get(self):
        user = User()
        request = RequestFactory().get('/something?abc=def')
        request.user = user
        criteria = QueryDictCriteriaWithRequester.build(request)
        self.assertEqual(request.user, criteria.user)
        self.assertDictEqual(request.GET, criteria.raw)
        file_request = FileRequest(requester=user, criteria=criteria.marshall())
        criteria = QueryDictCriteriaWithRequester.restore_from(file_request)
        self.assertEqual(request.user, criteria.user)
        self.assertDictEqual(request.GET, criteria.raw)

    def test_post(self):
        user = User()
        request = RequestFactory().post('/something?abc=def')
        request.user = user
        criteria = QueryDictCriteriaWithRequester.build(request)
        self.assertEqual(request.user, criteria.user)
        self.assertDictEqual(request.POST, criteria.raw)
        file_request = FileRequest(requester=user, criteria=criteria.marshall())
        criteria = QueryDictCriteriaWithRequester.restore_from(file_request)
        self.assertEqual(request.user, criteria.user)
        self.assertDictEqual(request.POST, criteria.raw)

    def test_delete(self):
        request = RequestFactory().delete('/something?abc=def')
        request.user = User()
        self.assertRaises(RuntimeError, QueryDictCriteriaWithRequester.build, request)


class QueryDictCriteriaTest(SimpleTestCase):

    def test_get(self):
        request = RequestFactory().get('/something?abc=def')
        criteria = QueryDictCriteria.build(request)
        self.assertEqual(request.GET, criteria.raw)
        criteria = QueryDictCriteria.restore_from(FileRequest(criteria=criteria.marshall()))
        self.assertEqual(request.GET, criteria.raw)

    def test_post(self):
        request = RequestFactory().post('/something?abc=def')
        criteria = QueryDictCriteria.build(request)
        self.assertEqual(request.POST, criteria.raw)
        criteria = QueryDictCriteria.restore_from(FileRequest(criteria=criteria.marshall()))
        self.assertEqual(request.POST, criteria.raw)


class CriteriaTest(SimpleTestCase):

    def test_not_implemented(self):
        self.assertRaises(NotImplementedError, Criteria({}).marshall)
        self.assertRaises(NotImplementedError, Criteria.restore_from, FileRequest(criteria='abc'))
        self.assertRaises(NotImplementedError, Criteria.build, RequestFactory().get('/something?abc=def'))


class SimplePatternExtension(SimplePattern):

    def __init__(self, dataset):
        self.dataset = dataset

    def build_dataset(self, http_request):
        return self.dataset

    def respond_with_file(self, http_request, peek_data):
        return 'file'

    def respond_with_delay(self, http_request, peek_data):
        return 'delay'


class SimplePatternTest(SimpleTestCase):

    def test(self):
        request = object()
        delayed, response = SimplePatternExtension(Dataset('abc', [], True)).respond_to(request)
        self.assertTrue(delayed)
        self.assertEqual('delay', response)
        delayed, response = SimplePatternExtension(Dataset('abc', [], False)).respond_to(request)
        self.assertFalse(delayed)
        self.assertEqual('file', response)

    def test_not_implemented(self):
        request = object()
        self.assertRaises(NotImplementedError, SimplePattern().build_dataset, request)
        self.assertRaises(NotImplementedError, SimplePattern().respond_with_file, request, {})
        self.assertRaises(NotImplementedError, SimplePattern().respond_with_delay, request, {})



# -*- coding: utf-8 -*-
from __future__ import unicode_literals, print_function, division, absolute_import
from django.core.exceptions import PermissionDenied, SuspiciousOperation
from django.core.signing import BadSignature, SignatureExpired
from django.http import Http404
from django.utils import six
import sys
from . import toolbox
from .models import FileRequest


class SuspiciousToken(SuspiciousOperation):
    """Exception to raise when the token is simply invalid.

    It subclasses SuspiciousOperation to take advantage of Django's reporting functionality,
    because An invalid token means either (a) the developer made a crucial mistake or (b) someone's
    trying to poke holes.

    """
    pass


class SignatureHasExpired(PermissionDenied):
    """Exception to raise when the signature was valid but has expired."""
    pass


class UserIsNotRequester(PermissionDenied):
    """Exception to raise when the current user is not the requester."""
    pass


def evaluate_request(request, token, require_requester=True, verify_requester=True, signer=None):
    """Utility method to evaluate a request before serving the file.

    This allows you to for example show a preview page to the user before serving the file.

    Note that the exception handling assumes you're using a Signer based on django.core.signing. It will turn:

    - a ``FileRequest.DoesNotExist`` into Django's `Http404` exception
    - a ``SignatureExpired`` into ``SignatureHasExpired``
    - a ``BadSignature`` into ``SuspiciousToken``

    If ``verify_requester`` is set to ``True`` (the default) and the current user does not match the requester, a
    ``UserIsNotRequester`` exception will be raised.

    :param request: the request
    :type request: django.http.request.HttpRequest
    :param token: the token
    :type token: str
    :param require_requester: whether we expect the token to contain the request
    :type require_requester: bool
    :param verify_requester: whether we need to verify the current user is the requester
    :type verify_requester: bool
    :param signer: signer to use
    :return: two-tuple containing the file request and the data in the token
    """
    file_request, data = None, None
    try:
        file_request, data = toolbox.decode(token, require_requester=require_requester, signer=signer)
    except FileRequest.DoesNotExist:
        raise Http404()
    except SignatureExpired as e:
        six.reraise(SignatureHasExpired, SignatureHasExpired(e), sys.exc_info()[2])
    except BadSignature as e:
        six.reraise(SuspiciousToken, SuspiciousToken(e), sys.exc_info()[2])
    if verify_requester and file_request.requester != request.user:
        raise UserIsNotRequester()
    return file_request, data


def serve_file(request, token, require_requester=True, verify_requester=True, signer=None):
    """Basic view to serve a file.

    Uses ``evaluate_request`` under the hood. Please refer to that function to view information about exceptions.

    :param request: the file request
    :type request: bgfiles.models.FileRequest
    :param token: the token
    :type token: str
    :param require_requester: whether we expect the token to contain the request
    :type require_requester: bool
    :param verify_requester: whether we need to verify the current user is the requester
    :type verify_requester: bool
    :param signer: signer to use
    :return: django.http.HTTPResponse

    """
    file_request, data = evaluate_request(request, token, require_requester=require_requester,
                                          verify_requester=verify_requester, signer=signer)
    return toolbox.serve(file_request)



# -*- coding: utf-8 -*-
from __future__ import unicode_literals, print_function, division, absolute_import
from contextlib import contextmanager
from django.utils import translation, timezone


handlers = {}


def register(handler_class, file_types):
    """Register a handler for one or more file types.

    :param handler_class: a pattern implementation
    :param file_types: list of supported file types
    """
    for file_type in file_types:
        handlers[file_type] = handler_class


def dispatch(file_request_id):
    """Dispatches the handling of the file request.

    This will try to retrieve the file request and use its file type to retrieve a registered handler.
    That handler will then be asked to process the request.

    :param file_request_id: id of the file request
    :return: bool indicating whether there was actually something processed
    """
    from ..models import FileRequest
    request = FileRequest.objects.select_for_update().get(id=file_request_id)
    if request.finished_at:
        return False
    handler_class = handlers[request.file_type]
    translation_ctx = translation.override if request.requester_language else dummy
    timezone_ctx = timezone.override if request.requester_timezone else dummy
    with translation_ctx(request.requester_language):
        with timezone_ctx(request.requester_timezone):
            handler_class.handle(request)
    return True


@contextmanager
def dummy(s):
    yield s
